from __future__ import annotations

from typing import Optional

import z3

z3False = z3.BoolVal(False)
z3True = z3.BoolVal(True)


def get_decls(f: z3.ExprRef, rs: Optional[dict[str, int]] = None) -> dict[str, int]:
    if rs is None:
        rs = {}
    if z3.z3_debug():
        assert z3.is_expr(f)
    if z3.is_const(f):
        if not z3.z3util.is_expr_val(f):
            rs.setdefault(str(f), 0)
        return rs
    if f.decl().kind() == z3.Z3_OP_UNINTERPRETED:
        rs.setdefault(str(f.decl()), f.decl().arity())
    for f_ in f.children():
        get_decls(f_, rs)
    return rs


def make_declare_funs(structs: dict[str, list[z3.ArithSortRef]]) -> str:
    return "\n".join(make_declare_fun(name, arity) for name, arity in structs.items())


def make_declare_fun(name, arity) -> str:
    args = " ".join("Int" for _ in range(arity))
    return f"(declare-fun {name} ({args}) Int)"


def z3And(a: z3.BoolRef, b: z3.BoolRef) -> z3.BoolRef:
    if z3.is_false(a) or z3.is_false(b):
        return z3False
    if z3.is_true(a):
        return b
    if z3.is_true(b):
        return a
    return z3.And(a, b)



import z3

AP_VAR_NAME = "ap"
PC_VAR_NAME = "pc"
FP_VAR_NAME = "fp"
MEMORY_MAP_NAME = "memory"
MEMORY_BOUND_VAR_NAME = "memory_bound"
PRIME_CONST_NAME = "prime"
IDENT_UF_NAME = "MEM"

ap, fp, pc, prime = z3.Ints(
    f"{AP_VAR_NAME} {FP_VAR_NAME} {PC_VAR_NAME} {PRIME_CONST_NAME}"
)
memory = z3.Function(MEMORY_MAP_NAME, z3.IntSort(), z3.IntSort())

HORUS_DECLS = {
    AP_VAR_NAME: ap,
    FP_VAR_NAME: fp,
    PC_VAR_NAME: pc,
    PRIME_CONST_NAME: prime,
    MEMORY_MAP_NAME: memory,
}



from __future__ import annotations

from typing import Optional

import lark
from starkware.cairo.lang.compiler.ast.expr import *
from starkware.cairo.lang.compiler.ast.expr import ExprIdentifier
from starkware.cairo.lang.compiler.error_handling import InputFile
from starkware.cairo.lang.compiler.parser_transformer import (
    ParserContext,
    ParserError,
    ParserTransformer,
)

import horus.compiler.parser
from horus.compiler.code_elements import (
    AnnotatedCodeElement,
    BoolConst,
    BoolExprAtom,
    BoolExprCompare,
    BoolNegation,
    BoolOperation,
    CodeElementAnnotation,
    CodeElementCheck,
    CodeElementLogicalVariableDeclaration,
    CodeElementStorageUpdate,
    ExprLogicalIdentifier,
)


class HorusTransformer(ParserTransformer):
    """
    Transforms lark trees generated by extended
    grammar `Horus.ebnf`.
    """

    def __init__(
        self,
        input_file: InputFile,
        parser_context: Optional[ParserContext],
        is_parsing_check: bool = False,
    ):
        self.is_parsing_check = is_parsing_check
        super().__init__(input_file, parser_context)

    @lark.v_args(meta=True)
    def commented_code_element(self, meta, value):
        comment: Optional[str] = (
            value[1][2:] if len(value) == 2 and value[1] is not None else None
        )

        if comment is not None:
            possible_annotation: str = comment.strip()
            if possible_annotation.startswith("@"):
                annotation_type = possible_annotation.split(" ")[0].strip()

                if annotation_type in [
                    "@pre",
                    "@post",
                    "@assert",
                    "@invariant",
                    "@declare",
                    "@storage_update",
                ]:
                    check = horus.compiler.parser.parse(
                        filename=self.input_file.filename,
                        code=possible_annotation,
                        code_type="annotation",
                        expected_type=CodeElementAnnotation,
                        parser_context=self.parser_context,
                    )

                    if isinstance(check, (CodeElementCheck, CodeElementStorageUpdate)):
                        check.unpreprocessed_rep = " ".join(
                            possible_annotation.split(" ")[1:]
                        )

                    code_elem = super().commented_code_element(meta, [value[0], None])
                    code_elem.code_elm = AnnotatedCodeElement(
                        check, code_elm=code_elem.code_elm
                    )
                    return code_elem
                else:
                    raise ParserError(
                        f"Unexpected annotation type: {annotation_type}",
                        location=self.meta2loc(meta),
                    )

        return super().commented_code_element(meta, value)

    @lark.v_args(meta=True)
    def logical_identifier(self, meta, value):
        identifier_name = ".".join(x.value for x in value)
        if value[0].value == "$Return":
            return ExprIdentifier(name=identifier_name, location=self.meta2loc(meta))
        else:
            return ExprLogicalIdentifier(
                name=identifier_name, location=self.meta2loc(meta)
            )

    @lark.v_args(meta=True)
    def logical_identifier_def(self, meta, value):
        if value[0].value == "$Return":
            raise ParserError(
                f"'$Return' is a reserved name.", location=self.meta2loc(meta)
            )
        return ExprLogicalIdentifier(name=value[0].value, location=self.meta2loc(meta))

    def atom_logical_identifier(self, value):
        return value[0]

    @lark.v_args(inline=True)
    def bool_formula_impl(self, lhs, rhs):
        return BoolOperation(lhs, rhs, "->")

    @lark.v_args(inline=True)
    def bool_formula_or(self, lhs, rhs):
        return BoolOperation(lhs, rhs, "|")

    @lark.v_args(inline=True)
    def bool_formula_and(self, lhs, rhs):
        return BoolOperation(lhs, rhs, "&")

    @lark.v_args(inline=True)
    def bool_unary_neg(self, formula):
        return BoolNegation(formula)

    @lark.v_args(inline=True)
    def bool_expr_true(self):
        return BoolConst(True)

    @lark.v_args(inline=True)
    def bool_expr_false(self):
        return BoolConst(False)

    @lark.v_args(inline=True)
    def bool_atom(self, expr):
        return BoolExprAtom(expr)

    @lark.v_args(inline=True, meta=True)
    def bool_expr_le(self, meta, lhs, rhs):
        return BoolExprCompare(lhs, rhs, "<=", location=self.meta2loc(meta))

    @lark.v_args(inline=True, meta=True)
    def bool_expr_lt(self, meta, lhs, rhs):
        return BoolExprCompare(lhs, rhs, "<", location=self.meta2loc(meta))

    @lark.v_args(inline=True, meta=True)
    def bool_expr_ge(self, meta, lhs, rhs):
        return BoolExprCompare(lhs, rhs, ">=", location=self.meta2loc(meta))

    @lark.v_args(inline=True, meta=True)
    def bool_expr_gt(self, meta, lhs, rhs):
        return BoolExprCompare(lhs, rhs, ">", location=self.meta2loc(meta))

    @lark.v_args(inline=True)
    def bool_expr_parentheses(self, formula):
        return formula

    @lark.v_args(inline=True, meta=True)
    def precond_annotation(self, meta, expr):
        return CodeElementCheck(
            CodeElementCheck.CheckKind.PRE_COND, expr, location=self.meta2loc(meta)
        )

    @lark.v_args(inline=True, meta=True)
    def postcond_annotation(self, meta, expr):
        return CodeElementCheck(
            CodeElementCheck.CheckKind.POST_COND, expr, location=self.meta2loc(meta)
        )

    @lark.v_args(inline=True, meta=True)
    def assert_annotation(self, meta, expr):
        return CodeElementCheck(
            CodeElementCheck.CheckKind.ASSERT, expr, location=self.meta2loc(meta)
        )

    @lark.v_args(inline=True, meta=True)
    def invariant_annotation(self, meta, expr):
        return CodeElementCheck(
            CodeElementCheck.CheckKind.INVARIANT, expr, location=self.meta2loc(meta)
        )

    @lark.v_args(inline=True, meta=True)
    def declare_annotation(self, meta, identifier, type):
        return CodeElementLogicalVariableDeclaration(
            identifier.name, type, location=self.meta2loc(meta)
        )

    @lark.v_args(inline=True, meta=True)
    def storage_update_annotation(self, meta, identifier, args, result):
        return CodeElementStorageUpdate(
            identifier.name, args, result, location=self.meta2loc(meta)
        )

    def transform(self, tree: lark.Tree):
        # The nodes of the tree imported from cairo.ebnf appear with
        # prefix starkware__cairo__lang__compiler__cairo__ which we remove here.
        for node in tree.iter_subtrees():
            node.data = node.data.replace(
                "starkware__cairo__lang__compiler__cairo__", ""
            )

        return super().transform(tree)



from starkware.cairo.lang.compiler.scoped_name import ScopedName

allowed_syscalls = [
    ScopedName.from_string("starkware.starknet.common.syscalls.get_caller_address"),
    ScopedName.from_string("starkware.starknet.common.syscalls.get_contract_address"),
    ScopedName.from_string("starkware.starknet.common.syscalls.get_block_timestamp"),
]



from __future__ import annotations

from importlib import resources
from typing import Optional, Tuple, Union

import lark
from lark.exceptions import LarkError, UnexpectedToken, VisitError
from lark.load_grammar import PackageResource
from starkware.cairo.lang.compiler.error_handling import InputFile, LocationError
from starkware.cairo.lang.compiler.parser import wrap_lark_error
from starkware.cairo.lang.compiler.parser_transformer import ParserContext

from horus.compiler.parser_transformer import HorusTransformer


def starkware_grammar_loader(
    base_path: Union[None, str, PackageResource], grammar_path: str
) -> tuple[str, str]:
    """
    A hack to load cairo.ebnf instead of cairo.lark.
    """
    if base_path is not None:
        raise IOError()  # lark import loading procedure will try
        # to use another loader in this case

    if not grammar_path.startswith("starkware"):
        raise IOError()  # same here

    actual_path = grammar_path.replace(".lark", ".ebnf")
    path_entries = actual_path.split("/")
    module_name = ".".join(path_entries[:-1])

    return (actual_path, resources.read_text(module_name, path_entries[-1]))


grammar = resources.read_text("horus.compiler", "horus.ebnf")
gram_parser = lark.Lark(
    grammar,
    start=[
        "cairo_file",
        "code_block",
        "code_element",
        "expr",
        "instruction",
        "type",
        "typed_identifier",
        "annotation",
    ],
    lexer="basic",
    parser="lalr",
    propagate_positions=True,
    import_paths=[starkware_grammar_loader],
)


def parse(
    filename: Optional[str],
    code: str,
    code_type: str,
    expected_type,
    parser_context: Optional[ParserContext] = None,
):
    """
    Copy-pasted function from StarkWare's compiler with only
    difference is that we use `HorusTransformer`
    instead of `ParserTranformer` and do some manipulations
    with logical identifier tokens.
    """
    is_parsing_check = code_type == "annotation"
    input_file = InputFile(filename=filename, content=code)
    parser_transformer = HorusTransformer(
        input_file, parser_context=parser_context, is_parsing_check=is_parsing_check
    )

    parser = gram_parser.parse_interactive(code, start=code_type)
    parser_state = parser.parser_state
    old_state_stack = list(parser_state.state_stack)
    old_value_stack = list(parser_state.value_stack)
    try:
        token = None
        for token in parser.lexer_state.lex(parser_state):
            if not is_parsing_check:
                # This is a hack to use the same grammar for
                # Cairo code and Horus checks.
                # If a Cairo code is being parsed
                # logical identifiers cannot appear anywhere.
                if token.type == "LOGICAL_IDENTIFIER":
                    accepts = parser.accepts()
                    accepts.remove("LOGICAL_IDENTIFIER")
                    raise UnexpectedToken(token=token, expected=accepts)  # type: ignore

            old_state_stack = list(parser_state.state_stack)
            old_value_stack = list(parser_state.value_stack)
            parser.feed_token(token)
        old_state_stack = list(parser_state.state_stack)
        old_value_stack = list(parser_state.value_stack)
        tree = parser.feed_eof(last_token=token)
    except UnexpectedToken as err:
        # Restore the old state stack.
        parser_state.state_stack = old_state_stack
        parser_state.value_stack = old_value_stack
        err.interactive_parser = parser
        raise wrap_lark_error(err, input_file) from None
    except LarkError as err:
        raise wrap_lark_error(err, input_file) from None

    try:
        parsed = parser_transformer.transform(tree)
    except VisitError as err:
        if isinstance(err.orig_exc, LocationError):
            raise err.orig_exc
        else:
            raise
    assert isinstance(
        parsed, expected_type
    ), f"Expected parsing result to be {expected_type.__name__}. Found: {type(parsed).__name__}"

    return parsed



from __future__ import annotations

import argparse
import dataclasses
import functools
import json
import os
import sys
import time
from typing import Callable, Dict, List, Optional, Sequence, Tuple

import starkware.cairo.lang.compiler.ast.visitor
import starkware.cairo.lang.compiler.parser
import starkware.cairo.lang.compiler.preprocessor.preprocess_codes
import starkware.cairo.lang.version
from starkware.cairo.lang.cairo_constants import DEFAULT_PRIME
from starkware.cairo.lang.compiler.ast.arguments import IdentifierList
from starkware.cairo.lang.compiler.ast.code_elements import CodeElementFunction
from starkware.cairo.lang.compiler.cairo_compile import (
    LIBS_DIR_ENVVAR,
    MAIN_SCOPE,
    START_FILE_NAME,
    generate_cairo_dependencies_file,
    get_codes,
    get_module_reader,
    get_start_code,
)
from starkware.cairo.lang.compiler.error_handling import LocationError
from starkware.cairo.lang.compiler.expression_transformer import ExpressionTransformer
from starkware.cairo.lang.compiler.identifier_manager import IdentifierManager
from starkware.cairo.lang.compiler.module_reader import ModuleReader
from starkware.cairo.lang.compiler.preprocessor.default_pass_manager import (
    PreprocessorStage,
)
from starkware.cairo.lang.compiler.preprocessor.identifier_aware_visitor import (
    IdentifierAwareVisitor,
)
from starkware.cairo.lang.compiler.preprocessor.pass_manager import (
    PassManager,
    PassManagerContext,
    Stage,
)
from starkware.cairo.lang.compiler.preprocessor.preprocessor import PreprocessedProgram
from starkware.cairo.lang.compiler.scoped_name import ScopedName
from starkware.starknet.compiler.compile import assemble_starknet_contract, get_abi
from starkware.starknet.compiler.starknet_pass_manager import starknet_pass_manager
from starkware.starknet.compiler.storage_var import STORAGE_VAR_DECORATOR
from starkware.starknet.compiler.validation_utils import has_decorator
from starkware.starknet.services.api.contract_class import ContractClass

import horus.compiler.parser
from horus.compiler.code_elements import AnnotatedCodeElement
from horus.compiler.contract_definition import HorusDefinition
from horus.compiler.preprocessor import HorusPreprocessor, HorusProgram


def assemble_horus_contract(
    preprocessed_program: HorusProgram, *args, **kwargs
) -> Tuple[ContractClass, HorusDefinition]:
    contract_definition = assemble_starknet_contract(
        preprocessed_program, *args, **kwargs
    )

    return (
        contract_definition,
        HorusDefinition(
            horus_version=horus.__version__,
            specifications=preprocessed_program.specifications,
            invariants=preprocessed_program.invariants,
            storage_vars=preprocessed_program.storage_vars,
        ),
    )


class HorusStorageVarDeclVisitor(IdentifierAwareVisitor):
    def __init__(
        self,
        storage_vars: Dict[ScopedName, IdentifierList],
        identifiers: Optional[IdentifierManager] = None,
    ):
        self.storage_vars = storage_vars
        super().__init__(identifiers)

    def _visit_default(self, obj):
        return obj

    def visit_CodeElementFunction(self, elm: CodeElementFunction):
        is_storage_var, storage_var_location = has_decorator(
            elm=elm, decorator_name=STORAGE_VAR_DECORATOR
        )
        if is_storage_var:
            self.storage_vars[self.current_scope + elm.name] = elm.arguments
        return elm


class HorusStorageVarCollectorStage(Stage):
    def run(self, context: PassManagerContext):
        assert isinstance(context, HorusPassManagerContext)
        visitor = HorusStorageVarDeclVisitor(
            storage_vars=context.storage_vars, identifiers=context.identifiers
        )
        for module in context.modules:
            visitor.visit(module)

        context.storage_vars = visitor.storage_vars
        return visitor


def horus_pass_manager(
    prime: int,
    read_module: Callable[[str], Tuple[str, str]],
    opt_unused_functions: bool = True,
    disable_hint_validation: bool = False,
) -> PassManager:

    manager = starknet_pass_manager(
        prime, read_module, opt_unused_functions, disable_hint_validation
    )
    manager.stages.insert(0, ("monkeypatch", MonkeyPatchStage()))
    preprocessor_stage = manager.stages[manager.get_stage_index("preprocessor")][1]
    manager.replace(
        "preprocessor",
        HorusPreprocessorStage(
            preprocessor_stage.prime,
            HorusPreprocessor,
            preprocessor_stage.auxiliary_info_cls,
            preprocessor_stage.preprocessor_kwargs,
        ),
    )
    manager.add_before(
        "storage_var_signature", "storage_collector", HorusStorageVarCollectorStage()
    )
    return manager


@dataclasses.dataclass
class HorusPassManagerContext(PassManagerContext):
    storage_vars: Dict[ScopedName, IdentifierList] = dataclasses.field(
        default_factory=dict
    )


class HorusPreprocessorStage(PreprocessorStage):
    def run(self, context: PassManagerContext):
        assert isinstance(context, HorusPassManagerContext)
        self.preprocessor_kwargs["storage_vars"] = context.storage_vars
        return super().run(context)


def preprocess_codes(
    codes: Sequence[Tuple[str, str]],
    pass_manager: PassManager,
    main_scope: ScopedName = ScopedName(),
    start_codes: Optional[List[Tuple[str, str]]] = None,
):
    """
    Preprocesses a list of Cairo files and returns a PreprocessedProgram instance.
    codes is a list of pairs (code_string, file_name).
    """
    context = HorusPassManagerContext(
        codes=list(codes),
        main_scope=main_scope,
        identifiers=IdentifierManager(),
        start_codes=[] if start_codes is None else start_codes,
    )

    pass_manager.run(context)

    assert context.preprocessed_program is not None
    return context.preprocessed_program


class MonkeyPatchStage(Stage):
    """
    Additional compilation stage we add before
    any other stage is performed in order to monkey-patch
    StarkWare definitions.
    All future monkey-patching should be done here.
    """

    def run(self, context: PassManagerContext):
        def visit_AnnotatedCodeElement(self, annotated_code_element):
            return AnnotatedCodeElement(
                annotation=annotated_code_element.annotation,
                code_elm=self.visit(annotated_code_element.code_elm),
            )

        starkware.cairo.lang.compiler.ast.visitor.Visitor.visit_AnnotatedCodeElement = (
            visit_AnnotatedCodeElement
        )
        starkware.cairo.lang.compiler.parser.parse = horus.compiler.parser.parse
        ExpressionTransformer.visit_ExprLogicalIdentifier = lambda self, expr: expr


def horus_compile_common(
    args: argparse.Namespace,
    pass_manager_factory: Callable[[argparse.Namespace, ModuleReader], PassManager],
    assemble_func: Callable,
) -> PreprocessedProgram:
    """
    Common code for CLI Cairo compilation.

    Arguments:
    args - Parsed arguments.
    pass_manager_factory - A pass manager factory.
    assemble_func - a function that converts a preprocessed program to the final output,
        the return value should be a Marshmallow dataclass.
    """

    start_time = time.time()
    debug_info = args.debug_info or args.debug_info_with_source

    try:
        codes = get_codes(args.files)
        out = args.output if args.output is not None else sys.stdout
        specs_out = args.spec_output if args.spec_output is not None else sys.stdout

        cairo_path: List[str] = list(
            filter(
                None,
                args.cairo_path.split(":") + os.getenv(LIBS_DIR_ENVVAR, "").split(":"),
            )
        )
        module_reader = get_module_reader(cairo_path=cairo_path)

        pass_manager = pass_manager_factory(args, module_reader)

        start_codes = []
        file_contents_for_debug_info = {}
        if getattr(args, "proof_mode", False):
            start_codes = [(get_start_code(), START_FILE_NAME)]
            file_contents_for_debug_info[START_FILE_NAME] = start_codes[0][0]

        preprocessed = preprocess_codes(
            codes=codes,
            pass_manager=pass_manager,
            main_scope=MAIN_SCOPE,
            start_codes=start_codes,
        )

        if args.preprocess:
            print(preprocessed.format(with_locations=debug_info), end="", file=out)
        else:
            if args.debug_info_with_source:
                for source_file in module_reader.source_files | set(args.files):
                    file_contents_for_debug_info[source_file] = open(source_file).read()

            assembled_program, specs = assemble_func(
                preprocessed,
                main_scope=MAIN_SCOPE,
                add_debug_info=debug_info,
                file_contents_for_debug_info=file_contents_for_debug_info,
            )

            json.dump(
                assembled_program.Schema().dump(assembled_program),
                out,
                indent=4,
                sort_keys=True,
            )
            # Print a new line at the end.
            print(file=out)

            json.dump(
                specs.Schema().dump(specs),
                specs_out,
                indent=4,
                sort_keys=True,
            )
            # Print a new line at the end.
            print(file=specs_out)

        return preprocessed
    finally:
        if args.cairo_dependencies:
            generate_cairo_dependencies_file(
                args.cairo_dependencies,
                module_reader.source_files | set(args.files),
                start_time,
            )


def cairo_compile_add_common_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        "files",
        metavar="file",
        type=str,
        nargs="+",
        help="One or more Cairo programs to compile.",
    )
    parser.add_argument(
        "--prime",
        type=int,
        default=DEFAULT_PRIME,
        help="The positive integer size of the finite field. This is a (usually large) prime power over which basic arithmetic within the program is carried out.",
    )
    parser.add_argument(
        "--cairo_path",
        type=str,
        default="",
        help=(
            'A list of directories, separated by ":" to resolve import paths. '
            "The full list will consist of directories defined by this argument, followed by "
            f"the environment variable {LIBS_DIR_ENVVAR}, the working directory and the standard "
            "library path."
        ),
    )
    parser.add_argument(
        "--preprocess",
        action="store_true",
        help="Stop after the preprocessor step and output the preprocessed program, which consists only of low-level Cairo (e.g. frame pointer and allocation pointer manipulations) along with annotations indicating relevant source code locations.",
    )
    parser.add_argument(
        "--output",
        type=argparse.FileType("w"),
        help="The output file name (default: stdout).",
    )
    parser.add_argument(
        "--no_debug_info",
        dest="debug_info",
        action="store_false",
        help="Don't include debug information in the compiled file. Removes the 'debug_info' field from the JSON output, which by default contains an 'instruction_locations' map with information on flow tracking data, hints, accessible scopes, and source code location.",
    )
    parser.add_argument(
        "--debug_info_with_source",
        action="store_true",
        help="Dump the source code of all relevant .cairo files into a 'file_contents' field in the 'debug_info' of the JSON output.",
    )
    parser.add_argument(
        "--cairo_dependencies",
        type=str,
        help="Path to dump a list of the Cairo source files used during the compilation as a CMake file.",
    )
    parser.add_argument(
        "--no_opt_unused_functions",
        dest="opt_unused_functions",
        action="store_false",
        default=True,
        help="Disable unused function optimization, which ordinarily only compiles functions reachable from the main scope in the dependency graph, i.e. functions that are actually called.",
    )


def main(args):
    parser = argparse.ArgumentParser(
        description="A tool to compile checked StarkNet contracts.",
        conflict_handler="resolve",
    )
    parser.add_argument(
        "--abi",
        type=argparse.FileType("w"),
        help=(
            "Dump the contract's ABI (application binary interface) to a file. "
            "This is a JSON list containing metadata (like type signatures and members) "
            "on functions, structs, and other things within the program."
        ),
    )
    parser.add_argument(
        "--disable_hint_validation",
        action="store_true",
        help="Disable the hint validation, which ordinarily checks program hints against a whitelist.",
    )
    parser.add_argument(
        "--account_contract",
        action="store_true",
        help="Compile as account contract, which means the ABI will be checked for expected builtin entry points.",
    )
    parser.add_argument(
        "--spec_output",
        type=argparse.FileType("w"),
        help="The specification output file name (default: stdout).",
    )

    def pass_manager_factory(
        args: argparse.Namespace, module_reader: ModuleReader
    ) -> PassManager:
        return horus_pass_manager(
            prime=args.prime,
            read_module=module_reader.read,
            opt_unused_functions=args.opt_unused_functions,
            disable_hint_validation=args.disable_hint_validation,
        )

    try:
        cairo_compile_add_common_args(parser)
        parser.add_argument(
            "-v",
            "--version",
            action="version",
            version=f"%(prog)s {horus.__version__}; cairo-compile {starkware.cairo.lang.version.__version__}",
        )
        args = parser.parse_args(args=args)
        assemble_func = functools.partial(
            assemble_horus_contract,
            filter_identifiers=False,
            is_account_contract=args.account_contract,
        )
        preprocessed = horus_compile_common(
            args=args,
            pass_manager_factory=pass_manager_factory,
            assemble_func=assemble_func,
        )
        abi = get_abi(preprocessed=preprocessed)
        if args.abi is not None:
            json.dump(abi, args.abi, indent=4, sort_keys=True)
            args.abi.write("\n")
    except LocationError as err:
        print(err, file=sys.stderr)
        return 1
    return 0


def run():
    main(sys.argv[1:])



from __future__ import annotations

import dataclasses
from enum import Enum
from typing import Optional, Sequence

from starkware.cairo.lang.compiler.ast.bool_expr import BoolEqExpr
from starkware.cairo.lang.compiler.ast.cairo_types import CairoType
from starkware.cairo.lang.compiler.ast.code_elements import CodeElement
from starkware.cairo.lang.compiler.ast.expr import ArgList, Expression
from starkware.cairo.lang.compiler.ast.formatting_utils import LocationField, Particle
from starkware.cairo.lang.compiler.ast.node import AstNode
from starkware.cairo.lang.compiler.error_handling import Location
from starkware.python.expression_string import ExpressionString


@dataclasses.dataclass
class ExprLogicalIdentifier(Expression):
    """
    Represents an expression of the form "$name".
    """

    name: str
    location: Optional[Location] = LocationField

    def to_expr_str(self) -> ExpressionString:
        return ExpressionString.highest(self.name)

    def get_children(self) -> Sequence[Optional["AstNode"]]:
        return []

    def get_particles(self) -> list[Particle]:
        return []


@dataclasses.dataclass
class BoolFormula(AstNode):
    """
    Base class for all boolean formula classes.
    """

    pass


@dataclasses.dataclass
class BoolExprAtom(BoolFormula):
    """
    Wrapper around Cairo's `BoolEqExpr`.
    """

    bool_expr: BoolEqExpr

    def get_children(self) -> Sequence[Optional["AstNode"]]:
        return [self.bool_expr]


@dataclasses.dataclass
class BoolExprCompare(BoolFormula):
    """
    A boolean formula of the form a <= b, a < b,
    a >= b or a > b.
    """

    a: Expression
    b: Expression
    comp: str
    location: Optional[Location] = LocationField

    def get_children(self) -> Sequence[Optional["AstNode"]]:
        return [self.a, self.b]


@dataclasses.dataclass
class BoolConst(BoolFormula):
    """
    A boolean constant `True` or `False`.
    """

    const: bool

    def get_children(self) -> Sequence[Optional["AstNode"]]:
        return []


@dataclasses.dataclass
class BoolOperation(BoolFormula):
    """
    Represents A and B, A or B, A -> B.
    """

    a: BoolFormula
    b: BoolFormula
    op: str

    def get_children(self) -> Sequence[Optional["AstNode"]]:
        return [self.a, self.b]


@dataclasses.dataclass
class BoolNegation(BoolFormula):
    """
    Represents ~A.
    """

    operand: BoolFormula

    def get_children(self) -> Sequence[Optional["AstNode"]]:
        return [self.operand]


class CodeElementAnnotation(CodeElement):
    pass


@dataclasses.dataclass
class CodeElementCheck(CodeElementAnnotation):
    """
    Represents a logical annotation of a kind
    specified by `check_kind`.
    """

    class CheckKind(Enum):
        POST_COND = "@post"
        PRE_COND = "@pre"
        ASSERT = "@assert"
        INVARIANT = "@invariant"

    check_kind: CheckKind
    formula: BoolFormula
    location: Optional[Location] = LocationField
    unpreprocessed_rep: str = ""

    def format(self, allowed_line_length):
        # TODO: implement better formatting
        return str(self.check_kind)

    def get_children(self) -> Sequence[Optional["AstNode"]]:
        return [self.formula]


@dataclasses.dataclass
class CodeElementLogicalVariableDeclaration(CodeElementAnnotation):
    """
    Represents a logical variable declaration.
    """

    name: str
    type: CairoType
    location: Optional[Location] = LocationField

    def format(self, allowed_line_length):
        return f"@declare {self.name}: {self.type.format()}"

    def get_children(self) -> Sequence[Optional["AstNode"]]:
        return []


@dataclasses.dataclass
class CodeElementStorageUpdate(CodeElementAnnotation):
    """
    Represents a storage variable change annotation.
    """

    name: str
    arguments: ArgList
    value: Expression
    location: Optional[Location] = LocationField
    unpreprocessed_rep: str = ""

    def format(self, allowed_line_length):
        return f"@storage_update {self.name}[{self.arguments.format()}] = {self.value.format()}"

    def get_children(self) -> Sequence[Optional["AstNode"]]:
        return [
            self.arguments,
            self.value,
        ]


@dataclasses.dataclass
class AnnotatedCodeElement(CodeElement):
    """
    Represents a code element with an annotation placed after.
    Note that empty lines are also code elements, so usually
    `code_elm` is a `CodeElementEmptyLine`.
    """

    annotation: CodeElementAnnotation
    code_elm: CodeElement

    def format(self, allowed_line_length):
        return f"{self.annotation.format(allowed_line_length)}\n{self.code_elm.format(allowed_line_length)}"

    def get_children(self) -> Sequence[Optional[AstNode]]:
        return [self.annotation, self.code_elm]



from __future__ import annotations

from dataclasses import field
from typing import Dict, List, Optional

import marshmallow.fields as mfields
import marshmallow_dataclass
import z3
from marshmallow.exceptions import ValidationError
from starkware.cairo.lang.compiler.ast.cairo_types import CairoType
from starkware.cairo.lang.compiler.fields import CairoTypeAsStr
from starkware.cairo.lang.compiler.scoped_name import ScopedName, ScopedNameAsStr

import horus
from horus.compiler.var_names import *
from horus.utils import z3And


class SexpField(mfields.Field):
    def _serialize(self, value: z3.ExprRef, attr, obj, **kwargs):
        return super()._serialize(value.sexpr().split("\n"), attr, obj, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs) -> z3.ExprRef:
        v = super()._deserialize(value, attr, data, **kwargs)
        ref_str = "\n".join(v)
        refs = z3.parse_smt2_string(ref_str, decls=HORUS_DECLS)
        if len(refs) != 1:
            raise ValidationError(f"Can't deserialize '{ref_str}'")
        return refs[0]  # type: ignore


@marshmallow_dataclass.dataclass
class StorageUpdate:
    arguments: List[z3.IntNumRef] = field(
        metadata=dict(marshmallow_field=mfields.List(SexpField())),
        default_factory=list,
    )
    value: z3.IntNumRef = field(
        metadata=dict(marshmallow_field=SexpField()), default=z3.IntVal(0)
    )
    source: str = field(metadata=dict(marshmallow_field=mfields.String()), default="")


@marshmallow_dataclass.dataclass(frozen=False)
class Annotation:
    sexpr: z3.BoolRef = field(
        metadata=dict(marshmallow_field=SexpField()), default=z3.BoolVal(True)
    )
    source: List[str] = field(
        metadata=dict(marshmallow_field=mfields.List(mfields.String())),
        default_factory=list,
    )

    def __and__(self, other):
        return Annotation(
            sexpr=z3And(self.sexpr, other.sexpr), source=self.source + other.source
        )


@marshmallow_dataclass.dataclass(frozen=False)
class FunctionAnnotations:
    pre: Annotation = field(
        metadata=dict(
            marshmallow_field=mfields.Nested(
                marshmallow_dataclass.class_schema(Annotation)
            )
        ),
        default=Annotation(),
    )
    post: Annotation = field(
        metadata=dict(
            marshmallow_field=mfields.Nested(
                marshmallow_dataclass.class_schema(Annotation)
            )
        ),
        default=Annotation(),
    )
    logical_variables: Dict[ScopedName, CairoType] = field(
        metadata=dict(
            marshmallow_field=mfields.Dict(ScopedNameAsStr(), CairoTypeAsStr())
        ),
        default_factory=dict,
    )
    decls: Dict[str, int] = field(
        metadata=dict(marshmallow_field=mfields.Dict(mfields.Str(), mfields.Int())),
        default_factory=dict,
    )
    storage_update: Dict[ScopedName, List[StorageUpdate]] = field(
        metadata=dict(
            marshmallow_field=mfields.Dict(
                ScopedNameAsStr(),
                mfields.List(
                    mfields.Nested(marshmallow_dataclass.class_schema(StorageUpdate))
                ),
            )
        ),
        default_factory=dict,
    )


@marshmallow_dataclass.dataclass(frozen=True)
class HorusDefinition:
    horus_version: Optional[str] = field(
        metadata=dict(marshmallow_field=mfields.String()), default=horus.__version__
    )
    specifications: Dict[ScopedName, FunctionAnnotations] = field(
        metadata=dict(
            marshmallow_field=mfields.Dict(
                ScopedNameAsStr(),
                mfields.Nested(marshmallow_dataclass.class_schema(FunctionAnnotations)),
            )
        ),
        default_factory=dict,
    )
    invariants: Dict[ScopedName, Annotation] = field(
        metadata=dict(
            marshmallow_field=mfields.Dict(
                ScopedNameAsStr(),
                mfields.Nested(marshmallow_dataclass.class_schema(Annotation)),
            )
        ),
        default_factory=dict,
    )
    storage_vars: Dict[ScopedName, int] = field(
        metadata=dict(marshmallow_field=mfields.Dict(ScopedNameAsStr(), mfields.Int())),
        default_factory=dict,
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional

from starkware.cairo.lang.compiler.ast.arguments import IdentifierList
from starkware.cairo.lang.compiler.ast.code_elements import (
    CodeBlock,
    CodeElement,
    CodeElementEmptyLine,
    CodeElementFunction,
    CodeElementLabel,
    CodeElementScoped,
)
from starkware.cairo.lang.compiler.error_handling import Location
from starkware.cairo.lang.compiler.identifier_definition import (
    FunctionDefinition,
    FutureIdentifierDefinition,
    LabelDefinition,
    StructDefinition,
)
from starkware.cairo.lang.compiler.identifier_manager import MissingIdentifierError
from starkware.cairo.lang.compiler.identifier_utils import get_struct_definition
from starkware.cairo.lang.compiler.scoped_name import ScopedName
from starkware.starknet.compiler.starknet_preprocessor import (
    StarknetPreprocessedProgram,
    StarknetPreprocessor,
)

from horus.compiler.code_elements import (
    AnnotatedCodeElement,
    CodeElementAnnotation,
    CodeElementCheck,
    CodeElementLogicalVariableDeclaration,
    CodeElementStorageUpdate,
)
from horus.compiler.contract_definition import (
    Annotation,
    FunctionAnnotations,
    StorageUpdate,
)
from horus.compiler.parser import *
from horus.compiler.z3_transformer import *
from horus.utils import get_decls

PRE_COND = CodeElementCheck.CheckKind.PRE_COND
POST_COND = CodeElementCheck.CheckKind.POST_COND
ASSERT = CodeElementCheck.CheckKind.ASSERT
INVARIANT = CodeElementCheck.CheckKind.INVARIANT


@dataclass
class HorusProgram(StarknetPreprocessedProgram):
    specifications: Dict[ScopedName, FunctionAnnotations]
    invariants: Dict[ScopedName, Annotation]
    storage_vars: Dict[ScopedName, int]


class HorusPreprocessor(StarknetPreprocessor):
    def __init__(self, **kwargs):
        self.storage_vars: Dict[ScopedName, IdentifierList] = kwargs.pop("storage_vars")
        super().__init__(**kwargs)
        self.specifications: Dict[ScopedName, FunctionAnnotations] = {}
        self.invariants: Dict[ScopedName, Annotation] = {}
        self.logical_identifiers: Dict[str, CairoType] = {}

        # This is used to defer pre/postcondition unfolding
        # until the visitor steps into the body of the function
        # when the preprocessor stumbles upon a function.
        self.current_checks: List[CodeElementAnnotation] = []
        self.current_function = None

        # Used for dummy labels
        self.current_fresh_index: int = 0

    def get_program(self) -> HorusProgram:
        starknet_program = super().get_program()

        for specification in self.specifications.values():
            specification.decls = {
                **get_decls(specification.pre.sexpr),
                **get_decls(specification.post.sexpr),
            }
            for var in HORUS_DECLS.keys():
                specification.decls.pop(var, None)

        storage_vars: Dict[ScopedName, int] = {}

        for storage_var, args in self.storage_vars.items():
            size = sum(
                (
                    self.get_size(arg.expr_type)
                    if arg.expr_type is not None
                    else self.get_size(TypeFelt())
                    for arg in args.identifiers
                )
            )

            storage_vars[storage_var] = size

        return HorusProgram(
            **starknet_program.__dict__,
            specifications=self.specifications,
            invariants=self.invariants,
            storage_vars=storage_vars,
        )

    def visit_CodeBlock(self, code_block: CodeBlock):
        return super().visit_CodeBlock(code_block)

    def add_dummy_label_with_assert(self, assrt: CodeElementCheck):
        if not isinstance(
            self.identifiers.get_by_full_name(self.current_scope),
            FunctionDefinition,
        ):
            raise (
                PreprocessorError(
                    f"Cannot use @assert annotation outside of a function.",
                    location=assrt.location,
                )
            )

        name = f"!anonymous_assert_label_{self.current_fresh_index}"
        self.current_fresh_index += 1
        self.identifiers.add_identifier(
            name=self.current_scope + name,
            definition=FutureIdentifierDefinition(identifier_type=LabelDefinition),
        )
        self.add_label(ExprIdentifier(name))
        z3_transformer = Z3Transformer(
            self.identifiers,
            self,
            self.logical_identifiers,
            self.storage_vars,
            is_post=False,
        )
        expr = z3_transformer.visit(assrt.formula)
        self.invariants[self.current_scope + name] = Annotation(
            sexpr=expr,
            source=[assrt.unpreprocessed_rep],
        )

    def visit_AnnotatedCodeElement(self, annotated_code_element: AnnotatedCodeElement):
        if isinstance(annotated_code_element.annotation, CodeElementCheck):
            if annotated_code_element.annotation.check_kind == ASSERT:
                self.add_dummy_label_with_assert(annotated_code_element.annotation)

                # Turning the function into annotated one.
                current_annotations = self.specifications.get(
                    self.current_scope, FunctionAnnotations()
                )
                self.specifications[self.current_scope] = current_annotations

                return self.visit(annotated_code_element.code_elm)

        result = self.visit(annotated_code_element.code_elm)
        self.current_checks.append(annotated_code_element.annotation)
        return result

    def visit(self, obj):
        unfolded_obj = obj.code_elm if isinstance(obj, AnnotatedCodeElement) else obj
        if self.current_checks:
            if not isinstance(unfolded_obj, CodeElementEmptyLine):
                if (
                    isinstance(unfolded_obj, CodeElementFunction)
                    and unfolded_obj.element_type == "func"
                ):
                    self.current_function = unfolded_obj
                elif isinstance(unfolded_obj, CodeElementScoped):
                    if (
                        isinstance(unfolded_obj.code_elements[0], CodeElementFunction)
                        and unfolded_obj.code_elements[0].element_type == "func"
                    ):
                        self.current_function = unfolded_obj.code_elements[0]
                else:
                    self.compile_annotations(unfolded_obj)

        return super().visit(obj)

    def add_logical_variable(
        self,
        declaration: CodeElementLogicalVariableDeclaration,
        is_member: bool = False,
    ):
        declaration.type = self.resolve_type(declaration.type)
        if isinstance(declaration.type, TypeStruct):
            definition = self.identifiers.get_by_full_name(declaration.type.scope)

            assert isinstance(
                definition, StructDefinition
            ), "TypeStruct must contain StructDefinition"

            for member_name, member_definition in definition.members.items():
                self.add_logical_variable(
                    CodeElementLogicalVariableDeclaration(
                        declaration.name + "." + member_name,
                        member_definition.cairo_type,
                    ),
                    is_member=True,
                )
        elif isinstance(declaration.type, TypeTuple):
            for i, member in enumerate(declaration.type.members):
                if member.name is not None:
                    member_name = f"{declaration.name}.{member.name}"
                else:
                    member_name = f"{declaration.name}.{i}"
                self.add_logical_variable(
                    CodeElementLogicalVariableDeclaration(
                        member_name, self.resolve_type(member.typ)
                    ),
                    is_member=True,
                )

        if not is_member:
            current_annotations = self.specifications.get(
                self.current_scope, FunctionAnnotations()
            )
            variables_of_the_function = current_annotations.logical_variables
            variables_of_the_function[
                ScopedName.from_string(declaration.name)
            ] = self.resolve_type(declaration.type)
            self.specifications[self.current_scope] = current_annotations

        self.logical_identifiers[declaration.name] = declaration.type

    def add_state_change(self, decl: CodeElementStorageUpdate):
        z3_transformer = Z3Transformer(
            self.identifiers,
            self,
            self.logical_identifiers,
            self.storage_vars,
            is_post=True,
        )
        z3_expr_transformer = Z3ExpressionTransformer(
            identifiers=self.identifiers, z3_transformer=z3_transformer
        )

        decl_full_name = self.identifiers.search(
            self.accessible_scopes, ScopedName.from_string(decl.name)
        ).canonical_name

        if not decl_full_name in self.storage_vars:
            raise PreprocessorError(
                f"{decl_full_name} is not a storage variable", location=decl.location
            )

        storage_var_args = get_struct_definition(
            decl_full_name + "read" + "Args", self.identifiers
        )
        assert isinstance(storage_var_args, StructDefinition)

        if len(storage_var_args.members) != len(decl.arguments.args):
            raise PreprocessorError(
                f"Incorrect number of arguments for a storage map.",
                location=decl.location,
            )

        for storage_def in storage_var_args.members.values():
            assert isinstance(
                storage_def.cairo_type, TypeFelt
            ), "Non-felt arguments of storage maps are not supported yet."

        args = []
        for [storage_arg_name, storage_def], arg in zip(
            storage_var_args.members.items(), decl.arguments.args
        ):
            if arg.identifier and arg.identifier.name != storage_arg_name:
                raise PreprocessorError(
                    f"Wrong argument name: {arg.identifier.name}",
                    location=decl.location,
                )

            arg_expr, arg_type = simplify_and_get_type(
                arg.expr,
                self,
                self.logical_identifiers,
                self.storage_vars,
                is_post=True,
            )

            if not isinstance(arg_type, TypeFelt):
                raise PreprocessorError(
                    f"Argument value must have type {TypeFelt().format}",
                    location=decl.location,
                )
            args.append(z3_expr_transformer.visit(arg_expr))

        current_annotations = self.specifications.get(
            self.current_scope, FunctionAnnotations()
        )
        storage_updates = current_annotations.storage_update.get(decl_full_name, [])
        storage_update = StorageUpdate(
            args,
            z3_expr_transformer.visit(
                simplify(
                    decl.value,
                    self,
                    self.logical_identifiers,
                    self.storage_vars,
                    is_post=True,
                )
            ),
            decl.unpreprocessed_rep,
        )
        storage_updates.append(storage_update)
        current_annotations.storage_update[decl_full_name] = storage_updates
        self.specifications[self.current_scope] = current_annotations

    def compile_annotations(self, code_elem: CodeElement):
        def append_check(
            check_kind: CodeElementCheck.CheckKind,
            key: Optional[ScopedName],
            check: Annotation,
        ):
            current_annotations = self.specifications.get(
                self.current_scope, FunctionAnnotations()
            )
            if check_kind == PRE_COND:
                current_annotations.pre = current_annotations.pre & check
            elif check_kind == POST_COND:
                current_annotations.post = current_annotations.post & check
            elif check_kind == INVARIANT:
                assert key is not None
                current_invariant = self.invariants.get(key, Annotation())
                self.invariants[key] = current_invariant & check

            self.specifications[self.current_scope] = current_annotations

        for parsed_check in self.current_checks:
            try:
                if (
                    isinstance(parsed_check, CodeElementCheck)
                    and parsed_check.check_kind == INVARIANT
                ):
                    if not isinstance(code_elem, CodeElementLabel):
                        raise PreprocessorError(
                            "@invariant annotation must be placed before a label",
                            parsed_check.location,
                        )
                elif isinstance(parsed_check, CodeElementCheck):
                    if not (
                        isinstance(code_elem, CodeElementFunction)
                        and code_elem.element_type == "func"
                    ):
                        raise PreprocessorError(
                            f"{parsed_check.check_kind} annotation is not allowed here",
                            parsed_check.location,
                        )

                if isinstance(parsed_check, CodeElementLogicalVariableDeclaration):
                    self.add_logical_variable(parsed_check)
                elif isinstance(parsed_check, CodeElementStorageUpdate):
                    self.add_state_change(parsed_check)
                elif isinstance(parsed_check, CodeElementCheck):
                    is_post = parsed_check.check_kind == POST_COND
                    z3_transformer = Z3Transformer(
                        self.identifiers,
                        self,
                        self.logical_identifiers,
                        self.storage_vars,
                        is_post,
                    )
                    expr = z3_transformer.visit(parsed_check.formula)
                    annotation = Annotation(
                        sexpr=expr, source=[parsed_check.unpreprocessed_rep]
                    )

                    if parsed_check.check_kind == INVARIANT:
                        append_check(
                            parsed_check.check_kind,
                            self.current_scope + code_elem.identifier.name,
                            annotation,
                        )
                    elif parsed_check.check_kind in (PRE_COND, POST_COND):
                        append_check(
                            parsed_check.check_kind,
                            None,
                            annotation,
                        )
            except MissingIdentifierError as e:
                raise PreprocessorError(str(e), location=parsed_check.location)

        self.current_checks = []

    def visit_function_body_with_retries(
        self, code_block: CodeBlock, location: Optional[Location]
    ):
        # This is needed because pre/postconditions can refer to arguments of the function.
        # So it's easier to process the conditions when the preprocessor
        # has stepped into the body of the function.

        if self.current_function is not None:
            self.compile_annotations(self.current_function)
            self.current_function = None
            self.current_checks = []

        super().visit_function_body_with_retries(code_block, location)
        self.logical_identifiers = {}



from __future__ import annotations

from typing import Dict, Optional

import z3
from starkware.cairo.lang.compiler.ast.arguments import IdentifierList
from starkware.cairo.lang.compiler.ast.bool_expr import BoolEqExpr
from starkware.cairo.lang.compiler.ast.cairo_types import (
    CairoType,
    TypeFelt,
    TypePointer,
    TypeStruct,
    TypeTuple,
)
from starkware.cairo.lang.compiler.ast.expr import (
    ArgList,
    ExprAddressOf,
    ExprCast,
    ExprConst,
    ExprDeref,
    ExprDot,
    Expression,
    ExprFutureLabel,
    ExprHint,
    ExprIdentifier,
    ExprNeg,
    ExprNewOperator,
    ExprOperator,
    ExprParentheses,
    ExprPow,
    ExprReg,
    ExprSubscript,
    ExprTuple,
)
from starkware.cairo.lang.compiler.ast.expr_func_call import ExprFuncCall
from starkware.cairo.lang.compiler.expression_simplifier import ExpressionSimplifier
from starkware.cairo.lang.compiler.identifier_definition import (
    FunctionDefinition,
    NamespaceDefinition,
    StructDefinition,
)
from starkware.cairo.lang.compiler.identifier_manager import IdentifierManager
from starkware.cairo.lang.compiler.identifier_utils import get_struct_definition
from starkware.cairo.lang.compiler.instruction import Register
from starkware.cairo.lang.compiler.preprocessor.identifier_aware_visitor import (
    IdentifierAwareVisitor,
)
from starkware.cairo.lang.compiler.preprocessor.preprocessor import Preprocessor
from starkware.cairo.lang.compiler.preprocessor.preprocessor_error import (
    PreprocessorError,
)
from starkware.cairo.lang.compiler.resolve_search_result import resolve_search_result
from starkware.cairo.lang.compiler.scoped_name import ScopedName
from starkware.cairo.lang.compiler.type_casts import CairoTypeError
from starkware.cairo.lang.compiler.type_system_visitor import *
from starkware.cairo.lang.compiler.type_system_visitor import simplify_type_system

from horus.compiler.allowed_syscalls import allowed_syscalls
from horus.compiler.code_elements import (
    BoolConst,
    BoolExprCompare,
    BoolFormula,
    BoolNegation,
    BoolOperation,
    ExprLogicalIdentifier,
)
from horus.compiler.type_checker import simplify, simplify_and_get_type
from horus.compiler.var_names import *
from horus.utils import z3And, z3True


class Z3ExpressionTransformer(IdentifierAwareVisitor):
    def __init__(
        self,
        identifiers: Optional[IdentifierManager] = None,
        z3_transformer: Optional[Z3Transformer] = None,
    ):
        self.prime = z3.Int(PRIME_CONST_NAME)
        self.memory = z3.Function(MEMORY_MAP_NAME, z3.IntSort(), z3.IntSort())
        self.z3_transformer = z3_transformer
        if z3_transformer is not None:
            self.storage_vars = z3_transformer.storage_vars
        super().__init__(identifiers)

    def visit_ExprConst(self, expr: ExprConst):
        return z3.IntVal(expr.val)

    def visit_ExprHint(self, expr: ExprHint):
        raise PreprocessorError(
            "Usage of hints in assertions is not allowed", location=expr.location
        )

    def visit_ExprLogicalIdentifier(self, expr: ExprLogicalIdentifier):
        return z3.Int(expr.name)

    def visit_ExprIdentifier(self, expr: ExprIdentifier):
        return ExprIdentifier(name=expr.name, location=expr.location)

    def visit_ExprFutureLabel(self, expr: ExprFutureLabel):
        raise PreprocessorError(
            "Usage of labels in assertions is not allowed", location=expr.location
        )

    def visit_ExprReg(self, expr: ExprReg):
        if expr.reg == Register.AP:
            return z3.Int(AP_VAR_NAME)
        else:
            return z3.Int(FP_VAR_NAME)

    def visit_ExprOperator(self, expr: ExprOperator):
        a = self.visit(expr.a)
        b = self.visit(expr.b)

        if expr.op == "+":
            return a + b
        elif expr.op == "-":
            return a - b
        elif expr.op == "*":
            return a * b
        elif expr.op == "/":
            return a / b

    def visit_ExprPow(self, expr: ExprPow):
        raise PreprocessorError("Non-constant powers in assertions are not supported.")

    def visit_ExprNeg(self, expr: ExprNeg):
        return -self.visit(expr.val)

    def visit_ExprParentheses(self, expr: ExprParentheses):
        return self.visit(expr.val)

    def visit_ExprDeref(self, expr: ExprDeref):
        return self.memory(self.visit(expr.addr))

    def visit_ExprSubscript(self, expr: ExprSubscript):
        raise PreprocessorError(
            "This was supposed to be unreachable.",
            location=expr.location,
        )

    def visit_ExprDot(self, expr: ExprDot):
        raise PreprocessorError(
            "This was supposed to be unreachable.",
            location=expr.location,
        )

    def visit_ExprAddressOf(self, expr: ExprAddressOf):
        raise PreprocessorError(
            "This was supposed to be unreachable.",
            location=expr.location,
        )

    def visit_ExprCast(self, expr: ExprCast):
        if isinstance(expr.dest_type, TypeStruct):
            assert self.identifiers is not None
            assert self.z3_transformer is not None
            search_result = self.identifiers.search(
                self.z3_transformer.preprocessor.accessible_scopes, expr.dest_type.scope
            )
            definition = resolve_search_result(search_result, self.identifiers)

            if isinstance(definition, NamespaceDefinition):
                if not search_result.canonical_name in self.z3_transformer.storage_vars:
                    raise PreprocessorError(
                        f"{expr.dest_type.scope} is not a storage var.",
                        location=expr.location,
                    )

                arg_struct_def = get_struct_definition(
                    search_result.canonical_name + "read" + "Args", self.identifiers
                )
                assert isinstance(arg_struct_def, StructDefinition)

                storage_var = z3.Function(
                    str(search_result.canonical_name),
                    *[z3.IntSort() for _ in arg_struct_def.members.values()],
                    z3.IntSort(),
                )

                assert isinstance(expr.expr, ExprTuple)

                args = [self.visit(arg.expr) for arg in expr.expr.members.args]

                return storage_var(*args)
            elif isinstance(definition, FunctionDefinition):
                if search_result.canonical_name in allowed_syscalls:
                    return z3.Int(expr.dest_type.scope.path[-1].replace("get_", "%"))

        inner_expr = self.visit(expr.expr)
        return inner_expr

    def visit_ExprFuncCall(self, expr: ExprFuncCall):
        """
        If a function call appears in an assertion it is expected to be a storage variable
        reference. Otherwise an exception is thrown.
        """
        assert self.z3_transformer is not None
        search_result = self.identifiers.search(
            self.z3_transformer.preprocessor.accessible_scopes,
            ScopedName.from_string(expr.rvalue.func_ident.name),
        )
        definition = resolve_search_result(search_result, self.identifiers)
        if search_result.canonical_name in allowed_syscalls:
            return z3.Int(search_result.canonical_name.path[-1].replace("get_", "%"))

        if isinstance(definition, NamespaceDefinition):
            if not search_result.canonical_name in self.z3_transformer.storage_vars:
                raise PreprocessorError(
                    f"{expr.rvalue.func_ident.name} is not a storage var.",
                    location=expr.location,
                )

            arg_struct_def = get_struct_definition(
                search_result.canonical_name + "read" + "Args", self.identifiers
            )
            assert isinstance(arg_struct_def, StructDefinition)

            storage_var = z3.Function(
                str(search_result.canonical_name),
                *[z3.IntSort() for _ in arg_struct_def.members.values()],
                z3.IntSort(),
            )

            args = [self.visit(arg.expr) for arg in expr.rvalue.arguments.args]

            return storage_var(*args)

        raise PreprocessorError(
            f"Function {search_result.canonical_name} cannot be used in assertions.",
            location=expr.location,
        )

    def visit_ArgList(self, arg_list: ArgList):
        raise PreprocessorError(
            "This was supposed to be unreachable.",
            location=arg_list.location,
        )

    def visit_ExprTuple(self, expr: ExprTuple):
        raise PreprocessorError(
            "This was supposed to be unreachable.",
            location=expr.location,
        )

    def visit_ExprNewOperator(self, expr: ExprNewOperator):
        raise PreprocessorError(
            "Usage of the new operator in assertions is not allowed",
            location=expr.location,
        )


def get_smt_expression(expr: Expression, identifiers: IdentifierManager):
    return Z3ExpressionTransformer(identifiers).visit(expr)


class Z3Transformer(IdentifierAwareVisitor):
    def __init__(
        self,
        identifiers: IdentifierManager,
        preprocessor: Preprocessor,
        logical_identifiers: Dict[str, CairoType],
        storage_vars: Dict[ScopedName, IdentifierList],
        is_post: bool = False,
    ):
        super().__init__(identifiers)
        self.preprocessor = preprocessor
        self.logical_identifiers = logical_identifiers
        self.is_post = is_post
        self.storage_vars = storage_vars
        self.z3_expression_transformer = Z3ExpressionTransformer(identifiers, self)

    def visit(self, formula: BoolFormula):
        funcname = f"visit_{type(formula).__name__}"
        return getattr(self, funcname)(formula)

    def get_element_at(self, expr: Expression, ind: int):
        if isinstance(expr, ExprTuple):
            member = expr.members.args[ind]
            return member.expr
        else:
            return ExprSubscript(expr, ExprConst(ind))

    def get_member(self, expr: Expression, name: str):
        if isinstance(expr, ExprTuple):
            for member in expr.members.args:
                if member.identifier is not None and member.identifier.name == name:
                    return simplify(
                        member.expr,
                        self.preprocessor,
                        self.logical_identifiers,
                        self.storage_vars,
                        self.is_post,
                    )

            raise PreprocessorError(f"No member with the name {name}")
        else:
            return ExprDot(expr, ExprIdentifier(name))

    def make_tuple_eq(self, a: Expression, b: Expression, type: TypeTuple):
        result = z3True
        for i, member in enumerate(type.members):
            if member.name is not None:
                member_a = self.get_member(a, member.name)
                member_b = self.get_member(b, member.name)
            else:
                member_a = self.get_element_at(a, i)
                member_b = self.get_element_at(b, i)

            if isinstance(member.typ, (TypeFelt, TypePointer)):
                result = z3And(
                    result,
                    get_smt_expression(
                        simplify(
                            member_a,
                            self.preprocessor,
                            self.logical_identifiers,
                            self.storage_vars,
                            self.is_post,
                        ),
                        self.identifiers,
                    )
                    == get_smt_expression(
                        simplify(
                            member_b,
                            self.preprocessor,
                            self.logical_identifiers,
                            self.storage_vars,
                            self.is_post,
                        ),
                        self.identifiers,
                    ),
                )
            elif isinstance(member.typ, TypeStruct):
                result = z3And(
                    result, self.make_struct_eq(member_a, member_b, member.typ)
                )
            elif isinstance(member.typ, TypeTuple):
                result = z3And(
                    result, self.make_tuple_eq(member_a, member_b, member.typ)
                )

        return result

    def make_struct_eq(
        self, a: Expression, b: Expression, type: TypeStruct
    ) -> z3.BoolRef:
        definition = get_struct_definition(
            struct_name=type.scope, identifier_manager=self.identifiers
        )
        assert isinstance(
            definition, StructDefinition
        ), "TypeStruct must contain StructDefinition"
        result = z3True
        for member_name, member_definition in definition.members.items():
            member_a = self.get_member(a, member_name)
            member_b = self.get_member(b, member_name)
            if isinstance(member_definition.cairo_type, (TypeFelt, TypePointer)):
                result = z3And(
                    result,
                    get_smt_expression(
                        simplify(
                            member_a,
                            self.preprocessor,
                            self.logical_identifiers,
                            self.storage_vars,
                            self.is_post,
                        ),
                        self.identifiers,
                    )
                    == get_smt_expression(
                        simplify(
                            member_b,
                            self.preprocessor,
                            self.logical_identifiers,
                            self.storage_vars,
                            self.is_post,
                        ),
                        self.identifiers,
                    ),
                )
            elif isinstance(member_definition.cairo_type, TypeStruct):
                result = z3And(
                    result,
                    self.make_struct_eq(
                        member_a, member_b, member_definition.cairo_type
                    ),
                )
            elif isinstance(member_definition.cairo_type, TypeTuple):
                result = z3And(
                    result,
                    self.make_tuple_eq(
                        member_a, member_b, member_definition.cairo_type
                    ),
                )
            else:
                raise NotImplementedError("test")

        return result

    def visit_BoolEqExpr(self, bool_expr: BoolEqExpr):
        a, a_type = simplify_and_get_type(
            bool_expr.a,
            self.preprocessor,
            self.logical_identifiers,
            self.storage_vars,
            self.is_post,
        )
        b, b_type = simplify_and_get_type(
            bool_expr.b,
            self.preprocessor,
            self.logical_identifiers,
            self.storage_vars,
            self.is_post,
        )

        if a_type != b_type:
            raise CairoTypeError(
                f"Types of lhs and rhs must coincide. Got {a_type.format()} and {b_type.format()}",
                bool_expr.location,
            )

        if isinstance(a_type, TypeStruct):
            result = self.make_struct_eq(bool_expr.a, bool_expr.b, a_type)
        elif isinstance(a_type, TypeTuple):
            result = self.make_tuple_eq(bool_expr.a, bool_expr.b, a_type)
        else:
            a = self.z3_expression_transformer.visit(a)
            b = self.z3_expression_transformer.visit(b)
            result = a == b

        if bool_expr.eq:
            return result
        else:
            return z3.Not(result)

    def visit_BoolExprCompare(self, formula: BoolExprCompare):
        a, a_type = simplify_and_get_type(
            formula.a,
            self.preprocessor,
            self.logical_identifiers,
            self.storage_vars,
            self.is_post,
        )
        b, b_type = simplify_and_get_type(
            formula.b,
            self.preprocessor,
            self.logical_identifiers,
            self.storage_vars,
            self.is_post,
        )

        if a_type != b_type:
            raise CairoTypeError(
                f"Types of lhs and rhs must coincide. Got {a_type.format()} and {b_type.format()}",
                formula.location,
            )

        if isinstance(a_type, (TypeFelt, TypePointer)):
            a = self.z3_expression_transformer.visit(a)
            b = self.z3_expression_transformer.visit(b)

            if formula.comp == "<=":
                return a <= b
            elif formula.comp == "<":
                return a < b
            elif formula.comp == ">=":
                return a >= b
            elif formula.comp == ">":
                return a > b
        else:
            raise CairoTypeError(
                f"Cannot compare values of type {a_type.format()}",
                location=formula.location,
            )

    def visit_BoolConst(self, formula: BoolConst):
        return z3.BoolVal(formula.const)

    def visit_BoolOperation(self, formula: BoolOperation):
        a = self.visit(formula.a)
        b = self.visit(formula.b)

        if formula.op == "&":
            return z3And(a, b)
        elif formula.op == "|":
            return z3.Or(a, b)
        elif formula.op == "->":
            return z3.Implies(a, b)
        else:
            raise PreprocessorError(f"unknown logical operation {formula.op}")

    def visit_BoolNegation(self, formula: BoolNegation):
        return z3.Not(self.visit(formula.operand))



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['horus', 'horus.compiler']

package_data = \
{'': ['*'], 'horus': ['scripts/*']}

install_requires = \
['cairo-lang==0.10.1',
 'eth-utils>=1.2.0,<2.0.0',
 'lark>=1.1.4,<2.0.0',
 'marshmallow-dataclass>=7.1.0,<8.5.4',
 'marshmallow>=3.15.0,<4.0.0',
 'z3-solver>=4.8.15,<5.0.0']

entry_points = \
{'console_scripts': ['horus-compile = horus.compiler.horus_compile:run']}

setup_kwargs = {
    'name': 'horus-compile',
    'version': '0.0.6.12',
    'description': 'Use formally verified annotations in your Cairo code',
    'long_description': 'This is the compiler for the Horus formal verification tool. See the [main\nrepository](https://github.com/NethermindEth/horus-checker) for full\ndocumentation.\n\n## Disclaimer\nKindly note, Horus is a tool consisting of two separate components: Horus-Check, released under the AGPLv3 license, and Horus-Compile, released under the Cairo Toolkit License. When "Horus" is referenced, the reference is to the two components jointly.\n\nHorus is currently in the alpha stage and no guarantee is being given as to the accuracy and/or completeness of any of the outputs the tool may generate. The tool is provided on an \'as is\' basis, without warranties or conditions of any kind, either express or implied, including without limitation as to the outputs of the verification process and the security of any system verified using Horus. As per the relevant licenses, to the fullest extent permitted by the law, Nethermind disclaims any liability in connection with your use of Horus and/or any of its outputs.\n\nPlease also note that the terminology used by Horus, including but not limited to words such as \'guarantee\', should be interpreted strictly within the remit of formal verification terminology. These words are not intended to, and shall not be construed as having legal significance of any kind.\n\nFor the avoidance of doubt, the outputs generated by Horus and/or your usage thereof shall not be considered or relied upon as any form of financial, investment, tax, legal, regulatory, or other advice.\n\nHorus-Check is licensed under AGPLv3 (Copyright (C) 2023 Nethermind). For more information on the dependencies, please see [here](https://github.com/NethermindEth/horus-checker/blob/master/LICENSE).\n\nHorus-Compile is licensed under the Cairo Toolkit License (Copyright (C) 2023 Nethermind), pursuant to an exception granted to Nethermind by Starkware Industries Ltd. For more information on the dependencies please see [here](https://github.com/NethermindEth/horus-compile/blob/master/LICENSE).',
    'author': 'Nethermind',
    'author_email': 'hello@nethermind.io',
    'maintainer': 'Ganesh Tiwari',
    'maintainer_email': 'ganesh@nethermind.io',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



from __future__ import annotations

import dataclasses
from typing import Dict, List, Optional, Tuple

from starkware.cairo.lang.compiler.ast.arguments import IdentifierList
from starkware.cairo.lang.compiler.ast.cairo_types import (
    CairoType,
    TypeFelt,
    TypeIdentifier,
    TypePointer,
    TypeStruct,
    TypeTuple,
)
from starkware.cairo.lang.compiler.ast.expr import (
    ExprAssignment,
    ExprCast,
    ExprConst,
    ExprDeref,
    ExprDot,
    Expression,
    ExprIdentifier,
    ExprOperator,
    ExprReg,
    ExprSubscript,
)
from starkware.cairo.lang.compiler.ast.expr_func_call import (
    ExprFuncCall,
    RvalueFuncCall,
)
from starkware.cairo.lang.compiler.expression_simplifier import ExpressionSimplifier
from starkware.cairo.lang.compiler.identifier_definition import (
    ConstDefinition,
    FunctionDefinition,
    NamespaceDefinition,
    ReferenceDefinition,
    TypeDefinition,
)
from starkware.cairo.lang.compiler.identifier_manager import (
    IdentifierManager,
    MissingIdentifierError,
)
from starkware.cairo.lang.compiler.identifier_utils import (
    get_struct_definition,
    get_type_definition,
)
from starkware.cairo.lang.compiler.instruction import Register
from starkware.cairo.lang.compiler.offset_reference import OffsetReferenceDefinition
from starkware.cairo.lang.compiler.preprocessor.preprocessor import Preprocessor
from starkware.cairo.lang.compiler.resolve_search_result import resolve_search_result
from starkware.cairo.lang.compiler.scoped_name import ScopedName
from starkware.cairo.lang.compiler.substitute_identifiers import (
    GetIdentifierCallback,
    GetStructMembersCallback,
    ResolveTypeCallback,
    SubstituteIdentifiers,
)
from starkware.cairo.lang.compiler.type_casts import CairoTypeError
from starkware.cairo.lang.compiler.type_system_visitor import TypeSystemVisitor
from starkware.crypto.signature.signature import FIELD_PRIME

from horus.compiler.allowed_syscalls import allowed_syscalls
from horus.compiler.code_elements import ExprLogicalIdentifier


def get_expr_addr(expr: Expression):
    if isinstance(expr, ExprLogicalIdentifier):
        return expr

    if not isinstance(expr, ExprDeref):
        raise CairoTypeError("Expression has no address.", location=expr.location)
    return expr.addr


class HorusTypeChecker(TypeSystemVisitor):
    def __init__(
        self,
        accessible_scopes: Optional[List[ScopedName]] = None,
        identifiers: Optional[IdentifierManager] = None,
        logical_identifiers: Dict[str, CairoType] = {},
        storage_vars: Dict[ScopedName, IdentifierList] = {},
    ):
        super().__init__(identifiers)
        self.accessible_scopes = accessible_scopes
        self.identifiers = identifiers
        self.logical_identifiers = logical_identifiers
        self.storage_vars = storage_vars

    def visit(self, expr: Expression) -> tuple[Expression, CairoType]:
        return super().visit(expr)  # type: ignore

    def visit_ExprLogicalIdentifier(
        self, expr: ExprLogicalIdentifier
    ) -> tuple[ExprLogicalIdentifier, CairoType]:
        return (expr, self.logical_identifiers[expr.name])

    def visit_ExprDot(self, expr: ExprDot):
        inner_expr, inner_type = self.visit(expr.expr)
        inner_type = self.resolve_type(inner_type)
        if isinstance(inner_expr, ExprLogicalIdentifier):
            if not isinstance(inner_type, (TypeStruct, TypeTuple)):
                raise CairoTypeError("wrong type", location=expr.location)

            struct_def = get_struct_definition(
                struct_name=inner_type.scope,
                identifier_manager=self.identifiers,
            )
            if expr.member.name not in struct_def.members:
                raise CairoTypeError(
                    f"Member '{expr.member.name}' does not appear in definition of struct "
                    f"'{inner_type.format()}'.",
                    location=expr.location,
                )

            return (
                ExprLogicalIdentifier(
                    f"{inner_expr.name}.{expr.member.name}", location=expr.location
                ),
                struct_def.members[expr.member.name].cairo_type,
            )
        else:
            return super().visit_ExprDot(expr)

    def visit_ExprSubscript(self, expr: ExprSubscript):
        inner_expr, inner_type = self.visit(expr.expr)
        offset_expr, offset_type = self.visit(expr.offset)

        if isinstance(inner_type, TypeTuple):
            if isinstance(inner_expr, ExprLogicalIdentifier):
                self.verify_offset_is_felt(offset_type, offset_expr.location)
                offset_expr = ExpressionSimplifier().visit(offset_expr)
                if not isinstance(offset_expr, ExprConst):
                    raise CairoTypeError(
                        "Subscript-operator for tuples supports only constant offsets, found "
                        f"'{type(offset_expr).__name__}'.",
                        location=offset_expr.location,
                    )
                offset_value = offset_expr.val

                tuple_len = len(inner_type.members)
                if not 0 <= offset_value < tuple_len:
                    raise CairoTypeError(
                        f"Tuple index {offset_value} is out of range [0, {tuple_len}).",
                        location=expr.location,
                    )
                logical_identifier_name = f"{inner_expr.name}.{offset_value}"
                return (
                    ExprLogicalIdentifier(logical_identifier_name),
                    self.logical_identifiers[logical_identifier_name],
                )

        return super().visit_ExprSubscript(expr)

    def visit_ExprCast(self, expr: ExprCast) -> Tuple[Expression, CairoType]:
        if isinstance(expr.dest_type, TypeStruct):
            assert self.identifiers is not None
            search_result = self.identifiers.search(
                self.accessible_scopes, expr.dest_type.scope
            )
            definition = resolve_search_result(search_result, self.identifiers)

            if isinstance(definition, NamespaceDefinition):
                try:
                    self.identifiers.search(
                        self.accessible_scopes, expr.dest_type.scope + "read"
                    )
                    self.identifiers.search(
                        self.accessible_scopes, expr.dest_type.scope + "write"
                    )

                    ret_type_def = get_type_definition(
                        search_result.canonical_name + "read" + "Return",
                        self.identifiers,
                    )
                    assert isinstance(ret_type_def, TypeDefinition)
                    assert isinstance(ret_type_def.cairo_type, TypeTuple)

                    if len(ret_type_def.cairo_type.members) != 1:
                        raise CairoTypeError(
                            "Storage maps with return tuple of length higher than 1 are not supported yet",
                            location=expr.location,
                        )

                    return expr, ret_type_def.cairo_type.members[0].typ
                except MissingIdentifierError:
                    # Failed to find the storage variable stuff.
                    raise CairoTypeError(
                        "Function calls are not allowed in assertions",
                        location=expr.location,
                    )
            elif isinstance(definition, FunctionDefinition):
                if search_result.canonical_name in allowed_syscalls:
                    return expr, TypeFelt(expr.location)

        return super().visit_ExprCast(expr)  # type: ignore

    def visit_ExprFuncCall(self, expr: ExprFuncCall) -> Tuple[ExprFuncCall, CairoType]:
        assert self.identifiers is not None
        search_result = self.identifiers.search(
            self.accessible_scopes, ScopedName.from_string(expr.rvalue.func_ident.name)
        )
        definition = resolve_search_result(search_result, self.identifiers)

        if search_result.canonical_name in allowed_syscalls:
            return expr, TypeFelt(expr.location)

        if isinstance(definition, NamespaceDefinition):
            try:
                self.identifiers.search(
                    self.accessible_scopes,
                    ScopedName.from_string(expr.rvalue.func_ident.name) + "read",
                )
                self.identifiers.search(
                    self.accessible_scopes,
                    ScopedName.from_string(expr.rvalue.func_ident.name) + "write",
                )

                ret_type_def = get_type_definition(
                    search_result.canonical_name + "read" + "Return",
                    self.identifiers,
                )
                assert isinstance(ret_type_def, TypeDefinition)
                assert isinstance(ret_type_def.cairo_type, TypeTuple)

                if len(ret_type_def.cairo_type.members) != 1:
                    raise CairoTypeError(
                        "Storage maps with return tuple of length higher than 1 are not supported yet",
                        location=expr.location,
                    )

                if len(expr.rvalue.arguments.args) != len(
                    self.storage_vars[search_result.canonical_name].identifiers
                ):
                    raise CairoTypeError(
                        f"Storage var {search_result.canonical_name} has {len(self.storage_vars[search_result.canonical_name].identifiers)} arguments. Provided {len(expr.rvalue.arguments.args)}",
                        location=expr.location,
                    )

                args_signature = self.storage_vars[
                    search_result.canonical_name
                ].identifiers.copy()
                is_named = False
                for arg in expr.rvalue.arguments.args:
                    assert isinstance(arg, ExprAssignment)

                    if arg.identifier is None:
                        if is_named:
                            raise CairoTypeError(
                                "Unnamed argument cannot follow named ones",
                                location=arg.location,
                            )

                        found_arg = args_signature.pop(0)
                    else:
                        is_named = True
                        found_arg = next(
                            (
                                x
                                for x in args_signature
                                if arg.identifier.name == x.identifier.name
                            ),
                            None,
                        )
                        if found_arg is None:
                            raise CairoTypeError(
                                f"Unknown argument {arg.identifier.name}",
                                location=arg.identifier.location,
                            )

                        args_signature.remove(found_arg)

                    _, arg_type = self.visit(arg.expr)

                    if arg_type != found_arg.expr_type:
                        raise CairoTypeError(
                            f"The argument is expected to have type {found_arg.expr_type}",
                            location=arg.expr.location,
                        )

                return expr, ret_type_def.cairo_type.members[0].typ
            except MissingIdentifierError as e:
                # Failed to find the storage variable stuff.
                raise CairoTypeError(
                    "Function calls are not allowed in assertions",
                    location=expr.location,
                ) from e

        raise CairoTypeError(
            "Function calls are not allowed in assertions",
            location=expr.location,
        )


def get_return_variable(name: str, preprocessor: Preprocessor):
    definition = get_type_definition(
        preprocessor.current_scope + "Return", preprocessor.identifiers
    )
    assert isinstance(definition, TypeDefinition)

    return_tuple = ExprCast(
        expr=ExprOperator(
            ExprReg(Register.AP),
            "-",
            ExprConst(preprocessor.get_size(definition.cairo_type)),
        ),
        dest_type=TypePointer(definition.cairo_type),
    )

    result = return_tuple
    for member_name in name.split("."):
        result = ExprDot(result, ExprIdentifier(member_name))

    return result


class HorusSubstituteIdentifiers(SubstituteIdentifiers):
    """
    Since storage vars are handled slightly different on the side of Horus
    we need to redefine the behaviour of substitution, which is done by
    this class.
    """

    def __init__(
        self,
        get_identifier_callback: GetIdentifierCallback,
        resolve_type_callback: ResolveTypeCallback = None,
        get_struct_members_callback: GetStructMembersCallback = None,
        identifiers: Optional[IdentifierManager] = None,
    ):
        super().__init__(
            get_identifier_callback, resolve_type_callback, get_struct_members_callback
        )
        self.identifiers = identifiers

    def visit_ExprFuncCall(self, expr: ExprFuncCall):
        try:
            rvalue = expr.rvalue
            _ = self.get_identifier_callback(rvalue.func_ident)

            new_args = []
            for arg in rvalue.arguments.args:
                assert isinstance(arg, ExprAssignment)
                new_expr = self.visit(arg.expr)
                new_args.append(dataclasses.replace(arg, expr=new_expr))

            expr.rvalue.arguments = dataclasses.replace(
                expr.rvalue.arguments, args=new_args
            )

            return expr
        except CairoTypeError:
            return super().visit_ExprFuncCall(expr)


def substitute_identifiers(
    expr: Expression,
    get_identifier_callback: GetIdentifierCallback,
    resolve_type_callback: ResolveTypeCallback = None,
    get_struct_members_callback: GetStructMembersCallback = None,
    identifiers: Optional[IdentifierManager] = None,
) -> Expression:
    """
    Replaces identifiers by other expressions according to the given callback.
    """
    return HorusSubstituteIdentifiers(
        get_identifier_callback=get_identifier_callback,
        resolve_type_callback=resolve_type_callback,
        get_struct_members_callback=get_struct_members_callback,
        identifiers=identifiers,
    ).visit(expr)


def simplify_and_get_type(
    expr: Expression,
    preprocessor: Preprocessor,
    logical_identifiers: Dict[str, CairoType],
    storage_vars: Dict[ScopedName, IdentifierList],
    is_post: bool,
) -> tuple[Expression, CairoType]:
    def get_identifier(expr: ExprIdentifier):
        if is_post:
            definition = get_struct_definition(
                preprocessor.current_scope + "ImplicitArgs", preprocessor.identifiers
            )

            if definition.members.get(expr.name.split(".")[0]) is not None:
                implicit_args_type = TypeStruct(
                    definition.full_name,
                )
                return_def = get_type_definition(
                    preprocessor.current_scope + "Return", preprocessor.identifiers
                )

                implicit_args_struct = ExprCast(
                    expr=ExprOperator(
                        ExprReg(Register.AP),
                        "-",
                        ExprConst(
                            definition.size
                            + preprocessor.get_size(return_def.cairo_type)
                        ),
                    ),
                    dest_type=TypePointer(implicit_args_type),
                )

                result = implicit_args_struct
                for member_name in expr.name.split("."):
                    result = ExprDot(result, ExprIdentifier(member_name))

                return result

        if expr.name.startswith("$Return."):
            return get_return_variable(expr.name[len("$Return.") :], preprocessor)

        search_result = preprocessor.identifiers.search(
            preprocessor.accessible_scopes,
            ScopedName.from_string(expr.name),
        )
        definition = resolve_search_result(search_result, preprocessor.identifiers)

        if isinstance(definition, ConstDefinition):
            return ExprConst(definition.value)

        if isinstance(definition, (ReferenceDefinition, OffsetReferenceDefinition)):
            return definition.eval(
                preprocessor.flow_tracking.reference_manager,
                preprocessor.flow_tracking.data,
            )

        if isinstance(definition, NamespaceDefinition):
            try:
                preprocessor.identifiers.get_by_full_name(
                    search_result.canonical_name + "read"
                )
                preprocessor.identifiers.get_by_full_name(
                    search_result.canonical_name + "write"
                )

                ret_type_def = get_type_definition(
                    search_result.canonical_name + "read" + "Return",
                    preprocessor.identifiers,
                )
                assert isinstance(ret_type_def, TypeDefinition)
                assert isinstance(ret_type_def.cairo_type, TypeTuple)

                if len(ret_type_def.cairo_type.members) != 1:
                    raise CairoTypeError(
                        "Storage maps with return tuple of length higher than 1 are not supported yet",
                        location=expr.location,
                    )

                return expr
            except MissingIdentifierError as e:
                # Failed to find the storage variable stuff.
                raise CairoTypeError(
                    "Function calls are not allowed in assertions",
                    location=expr.location,
                ) from e

        raise CairoTypeError(
            f'Cannot obtain identifier "{expr.name}". Expected a reference but got "{definition.TYPE}"',
            location=expr.location,
            notes=[
                "\033[33mhint: Did you try to reference a local variable in a '@pre' condition?",
                "hint: Local variables cannot be referenced in '@pre' or '@post'.",
                "hint:",
                "hint: Try using an '@assert' within the function body.\033[0m",
            ],
        )

    expr = substitute_identifiers(
        expr,
        get_identifier,
        preprocessor.resolve_type,
        identifiers=preprocessor.identifiers,
    )
    expr, expr_type = HorusTypeChecker(
        preprocessor.accessible_scopes,
        preprocessor.identifiers,
        logical_identifiers,
        storage_vars,
    ).visit(expr)
    expr_type = preprocessor.resolve_type(expr_type)
    expr = ExpressionSimplifier(prime=FIELD_PRIME).visit(expr)

    return (expr, expr_type)


def simplify(
    expr: Expression,
    preprocessor: Preprocessor,
    logical_identifiers: Dict[str, CairoType],
    storage_vars: Dict[ScopedName, IdentifierList],
    is_post: bool,
) -> Expression:
    return simplify_and_get_type(
        expr, preprocessor, logical_identifiers, storage_vars, is_post
    )[0]
